// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "omdistortimage.h"
#include "Dal.h"
#include<vector>
#include<iostream>
#include"utility.h"
#include "interpolation.h"



#include"f90cxx.h"
#include <cmath>

#include "CalServer.h"


/************************************************************************************
*
*
* Fortran interfaces
*
*
************************************************************************************/
F90MODPROCPROTO( void, omdistortion_mod, initialisedistortion2)(double *, double *, int *, int *);
F90MODPROCPROTO( void, omdistortion_mod, adddistortion)(double *, double *, double *, double *);
F90MODPROCPROTO( void, omdistortion_mod, removedistortion)(double *, double *, double *, double *);

#include "omwcs.h"

void initialise(string file, OMWCS &wcs);
void distortImage(OMWCS &wcs,
                  vector<vector<float> > &image, 
                  vector<vector<float> > &exposureImage, 
                   vector<vector<int16> > &qualityImage);

void createOutputImage(vector<vector<float> > &image, 
                       vector<vector<float> > &exposureImage, 
                       vector<vector<int16> > &qualityImage, OMWCS &wcs);

void setQualityImage(vector<vector<float> > &image, vector<vector<int16> > &qualityImage);

/*********************************************************************
*
*
*
*********************************************************************/
void omdistortimage()
{

    string file = stringParameter("set");


    DataSet * set  = dataSetServer->open(file, Dal::Read);


    int nx = 0, ny = 0;
    vector<vector<float> >  image = loadFloatImage( set, "PRIMARY", nx, ny);


    vector<vector<float> >  exposureImage(nx, vector<float>(ny, 0.0)); 

    if( set->hasBlock("EXPOSURE") )
    {

        exposureImage = loadFloatImage( set, "EXPOSURE", nx, ny);
    }


    vector<vector<int16> >  qualityImage = loadInt16Image( set, "QUALITY", nx, ny);


    OMWCS wcs(1);
    initialise( file, wcs );


    distortImage( wcs, image, exposureImage, qualityImage);
    createOutputImage(image, exposureImage, qualityImage, wcs);


}

/******************************************************************
*
*
*
******************************************************************/
void initialise(string file, OMWCS &wcs)
{


    wcs.data->load(file);
    string filter = wcs.data->filter;

      
    double x0 = wcs.data->crpix1;    // WINDOWX0
    double y0 = wcs.data->crpix2;    // WINDOWY0
 
    /****************************************************	
    *
    *	create an OM engine
    *
    ****************************************************/
    calServer.setInstrument(Xmm::OM);
    if(filter == "UVW2" || filter == "uvw2")
        calServer.state()->filter()->set(Xmm::UVW2);
    else if(filter == "UVM2" || filter == "uvm2")
        calServer.state()->filter()->set(Xmm::UVM2);
    else if(filter == "UVW1" || filter == "uvw1")
        calServer.state()->filter()->set(Xmm::UVW1);
    else if(filter == "U" || filter == "u")
        calServer.state()->filter()->set(Xmm::U);
    else if(filter == "B" || filter == "b")
        calServer.state()->filter()->set(Xmm::B);
    else if(filter == "V" || filter == "v")
        calServer.state()->filter()->set(Xmm::V);

    else if(filter == "GRISM1" || filter == "GRISM1")
        calServer.state()->filter()->set(Xmm::GRISM1);



    /***********************************************************************
    *
    * Initialise the distortion (mssllib)
    *
    ***********************************************************************/
    F90MODPROC( omdistortion_mod, initialisedistortion2)(&x0, &y0, 
               &wcs.data->binAxis, &wcs.data->binAxis);

 

}
/******************************************************************
*
*
*
*
******************************************************************/
void distortImage(OMWCS &wcs,
                  vector<vector<float> >   &image, 
                  vector<vector<float> >  &exposureImage, 
                  vector<vector<int16> > &qualityImage)
{


    int nx = image.size();
    int ny = image[0].size();

    double minX = 10000.0;
    double maxX = -minX;
    double minY = 10000.0;
    double maxY = -minY;

   

    double x0 = wcs.data->crval1;
    double y0 = wcs.data->crval2;

 
    for(int i = 0; i <= nx; i++)
    {

        double x = x0 + double(i);
        for(int j = 0; j <= ny; j++)
        {
            double y = y0 + double(j);
            double x1 = 0.0, y1 = 0.0;
            F90MODPROC(omdistortion_mod, adddistortion)(&x, &y, &x1, &y1);
            minX = min(minX, x1);
            maxX = max(maxX, x1);
            minY = min(minY, y1);
            maxY = max(maxY, y1);  
        }
    }


    //    minX = wcs.data->crpix1;
    //    minY = wcs.data->crpix2;
 
    cout << "x0=" << x0 << " y0=" << y0 << endl;

    cout << "Line 176 minx=" << minX << " maxX=" << maxX << endl;
    cout << "Line 177 miny=" << minY << " maxY=" << maxY << endl;


    /****************************************************************
    *
    * Get the dimensions of the undistorted image
    *
    *****************************************************************/

    //    minX = floor(minX - 1.0);
    //   minY = floor(minY - 1.0);
    //   maxX = ceil(maxX + 1.0);
    //  maxY = ceil(maxY + 1.0);

    int nx1 = (int)ceil(maxX - minX + 0.00001);
    int ny1 = (int)ceil(maxY - minY + 0.00001);


    cout << "nx1=" << nx1 << " ny1=" << ny1 << endl;



    vector<vector<float> > distortedImage( nx1, vector<float>(ny1, -1.0) );
    vector<vector<float> > distortedExposureImage( nx1, vector<float>(ny1, -1.0) );
    vector<vector<int16> > distortedQualityImage( nx1, vector<int16>(ny1, 0) );
  

    double minX1 = 10000.0;
    double maxX1 = -minX1;
    double minY1 = 10000.0;
    double maxY1 = -minY1;




    /*************************************************************
    *
    *
    * Set up the interpolants
    *
    *
    *************************************************************/
    Interpolation<float> interpolator1(image);
    Interpolation<float> interpolator2(exposureImage);
    Interpolation<int16> interpolator3(qualityImage);


    for(int i = 0; i < nx1; i++)
    {

        double x = minX + double(i + 1);
        
        for(int j = 0; j < ny1; j++)
        {

            double y = minY + double(j + 1);
            double x1 = 0.0, y1 = 0.0;
            F90MODPROC(omdistortion_mod, adddistortion)(&x, &y, &x1, &y1);
  
            float f = interpolator1.BilinearInterpolation(x1, y1);
            distortedImage[i][j] = f;
            f = interpolator2.BilinearInterpolation(x1, y1);
            distortedExposureImage[i][j] = f;

            int16 f1 = interpolator3.BilinearInterpolation(x1, y1);
     
            distortedQualityImage[i][j] = f1;

            minX1 = min(minX1, x1);
            maxX1 = max(maxX1, x1);
            minY1 = min(minY1, y1);
            maxY1 = max(maxY1, y1);
	  
          
        }
    }

 

    image = distortedImage;   
    exposureImage = distortedExposureImage;   

    qualityImage = distortedQualityImage;

}

/***************************************************************************
*
*
**************************************************************************/
void createOutputImage(vector<vector<float> > &image,
                       vector<vector<float> > &exposureImage, 
                       vector<vector<int16> > &qualityImage, 
                       OMWCS &wcs)
{


    string fileName = stringParameter("outset");

    string file1 = stringParameter("set");
    dataSetServer->copy( file1, fileName);

    DataSet * set  = dataSetServer->open(fileName, Dal::Modify);



    /***************************************************************
    *
    * Delete the PRIMARY and QUALITY blocks in the file
    *
    ****************************************************************/


    set->deleteBlock("PRIMARY");


  

    if(set->hasBlock("QUALITY") )
    {
        set->deleteBlock("QUALITY");    
 
    }


    int nx = (int)image.size();
    int ny = (int)image[0].size();
  

    /***********************************************************
    *
    * Add the PRIMARY image
    **********************************************************/
    addBlock(set, "PRIMARY", image, 0);





    if(set->hasBlock("EXPOSURE") )
    {
        set->deleteBlock("EXPOSURE");    
 
 

        /***********************************************************
        *
        * Add the EXPOSURE image
        *
        **********************************************************/
        addBlock(set, "EXPOSURE", exposureImage, 1);
    }

    /***********************************************************
    *
    * Add the quality image
    *
    ************************************************************/

    setQualityImage(image, qualityImage);
    addBlock(set, "QUALITY", qualityImage, 2);




    /*******************************************************************
    *
    *
    *******************************************************************/
    
    *set->addAttribute("DATAMODE", "") = "IMAGING_UNDISTORTED";



    *set->addAttribute("WINDOWX0", "") = wcs.data->crpix1;
    *set->addAttribute("WINDOWDX", "") = nx * wcs.data->binAxis;
    *set->addAttribute("WINDOWY0", "") = wcs.data->crpix2;
    *set->addAttribute("WINDOWDY", "") = ny * wcs.data->binAxis;

  

    dataSetServer->close(set); // Finished with the file
    


}
/*************************************************************
*
*
*
*
*
*************************************************************/
void setQualityImage(vector<vector<float> > &image, vector<vector<int16> > &qualityImage)
{
    int nx = image.size();
    int ny = image[0].size();

    for(int i = 0; i < nx; i++)
    {
    
        for(int j = 0; j < ny; j++)
        {
	    if(isNull(image[i][j]) || image[i][j] < 0.0) //  || bitSet(qualityImage[i][j], 0) )
	    {
	      qualityImage[i][j] = 1;
            }
            else
            {
	      qualityImage[i][j] = 0;
            }
        

        }
    }
}
