// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef OMWCS_H
#define OMWCH_H
#include<iostream>
#include <vector>
#include <map>
#include "astrometrydefs.h"

class DataSet;





/************************************************************************
*
*
*
************************************************************************/



class Astrometry {
public:
Astrometry() : _poscorrOkay(false), _posintrp(false), 
               _raOffset(0.0), _raOffsetError(0.0),
               _decOffset(0.0), _decOffsetError(0.0),
               _rmsResidual(0.0), _nMatches(0) {};
                
Astrometry(bool poscorrOkay, bool posintrp, 
                double raOffset, double raOffsetError,
                double decOffset, double decOffsetError, 
                double rmsResidual, int nMatches) : 
                     _poscorrOkay(poscorrOkay), _posintrp(posintrp), 
                     _raOffset(raOffset), _raOffsetError(raOffsetError),
                     _decOffset(decOffset), _decOffsetError(decOffsetError),
                     _rmsResidual(rmsResidual), _nMatches(nMatches) {};
                

    Astrometry( const Astrometry &c)
    {


        if (this == &c) return;

        _poscorrOkay = c._poscorrOkay;
        _posintrp = c._posintrp;
        _raOffset = c._raOffset;
        _raOffsetError = c._raOffsetError;

        _decOffset = c._decOffset;
        _decOffsetError = c._decOffsetError;

        _rmsResidual = c._rmsResidual;
        _nMatches = c._nMatches;

    };
    /******************************************************************
    *
    *
    *
    *
    *******************************************************************/
    Astrometry& operator=(const Astrometry& c)
    {

        if(this == &c) 
        {  
            return *this;
        }
 
        _poscorrOkay = c._poscorrOkay;
        _posintrp = c._posintrp;
        _raOffset = c._raOffset;
        _raOffsetError = c._raOffsetError;

        _decOffset = c._decOffset;
        _decOffsetError = c._decOffsetError;

        _rmsResidual = c._rmsResidual;
        _nMatches = c._nMatches;

        return *this;
    };

                     
    bool _poscorrOkay, _posintrp;
    double _raOffset, _raOffsetError;
    double _decOffset, _decOffsetError;
    double _rmsResidual;
    int _nMatches;


};
/*********************************************************************
*
*
*
********************************************************************/
class ImageData
{
public:
    int nx, ny;
    int numberGoodPixels;

    vector<vector<float>  > image;        // The image
    vector<vector<float> > backgroundImage;
    vector<double> x, y;
    vector<double> f;
    vector<double> fSquared;
    string fileName;   
    ImageData() : nx(0), ny(0), numberGoodPixels(0) {};

};


/******************************************************
*
* The base class
*
******************************************************/
class OMWCS;
class WcsData
{
 public:

    
    int nx, ny, binAxis;
    bool binBpe;

    double cdelt1;  // Ra plate-scale  (radians) 
    double cdelt2;  // Dec plate-scale (radians)
    double crpix1;  // Reference x pixel 
    double crpix2;  // Reference y pixel 
    double crval1;  // RA at crpix1 
    double crval2;  // Dec at crpix2 
    double _cdecz, _sdecz;
    double _raPnt;
    double _decPnt;
    double _paPnt;
    double _mjdMid;
   
    Astrometry _astrometry; 


    double xLeft, xRight;
    double yBottom, yTop;
    double _area;
    
    double _dx, _dy;    
  
    double deadFrac;
    double frameTime;
    float exposure;
    string filter;
    string object;
    string startObs;
    string endObs; 
    string _obsId;


    WcsData() : nx(0), ny(0), binAxis(0), binBpe(false), 
            cdelt1(0.0), cdelt2(0.0), 
            crpix1(0.0), crpix2(0.0), 
            crval1(0.0), crval2(0.0),
            _cdecz(0.0), _sdecz(0.0),
            _raPnt(0.0), _decPnt(0.0), 
            _paPnt(0.0), _mjdMid(0.0),
    
           _astrometry(), 


            xLeft(0.0), xRight(0.0),
                
            yBottom(0.0), yTop(0.0),
            _area(0.0),             
                 _dx(0.0), _dy(0.0),
            deadFrac(0.0), frameTime(0.0), 
            exposure(0.0),
            filter("UNKNOWN"), object("UNKNOWN"),
            startObs(""), endObs(""), _obsId("")
            {};


  
    virtual ~WcsData() {};  










    virtual bool load(string ) = 0;
    virtual void convert() = 0;
    //    virtual void setImageOffsets( OMWCS &c1, OMWCS &c2) = 0;
    virtual void setImageParameters() = 0;
    virtual void getCornerCoords() = 0;

    /**************************************************************************
    * Convert image 1 coordinates  x1, y1 on image 1 to image 2 coordinates x2, y2  
    *
    ***************************************************************************/

    virtual void convertX1Y1ToX2Y2(double xIn, double yIn, OMWCS &c1, double &xOut, double &yOut) = 0;

 

    virtual bool withinImage(double x, double y) = 0;
    // 2014-02-02 vny adding max frame taime and dead frac
    //virtual void createOutputFile(string fileName, vector< vector<real32> > &image, 
    //                                   vector< vector<real32> > &exposureImage,
    //                                   vector< vector<int16> > &qualityImage, vector<string> &fileList) = 0;


    virtual void createOutputFile(string fileName, vector< vector<real32> > &image, 
				  vector< vector<real32> > &exposureImage,
				  vector< vector<int16> > &qualityImage, vector<string> &fileList,
				  double deadFracMax, double frameTimeMax) = 0;


    virtual void convertToLowResImage(vector< vector<float> > &image, 
                vector< vector<float> > &exposureImage,
		vector< vector<int16> > &qualityImage) = 0;




    virtual void correctReferencePixel(double dx, double dy) = 0;
    virtual void correctReferencePixel() = 0;

    virtual void updateReferencePixelCorrection(OMWCS &c, double dx, double dy) = 0;
 

    virtual double distance(double x1, double y1, OMWCS &c, double &x2, double &y2) = 0;

    virtual double overlapFraction(OMWCS &c2) = 0;


    void setDates(vector<string> &fileList);

    void setPaPnt(double paPnt){_paPnt=paPnt;};

}; 
/******************************************************
*
*
******************************************************/
class OMWCS;
class WcsData1 : public WcsData
{
public:
  
    WcsData1() : WcsData() { };
    WcsData1& operator=(const WcsData1& rhs);

    ~WcsData1()  {};
    virtual bool load(string );
    virtual void convert();
    //   virtual void setImageOffsets( OMWCS &c1, OMWCS &c2);
    virtual void setImageParameters();
    virtual void getCornerCoords();
    inline virtual void convertX1Y1ToX2Y2(double xIn, double yIn, OMWCS &c1, double &xOut, double &yOut);
 
    virtual bool withinImage(double x, double y);
    virtual void createOutputFile(string fileName, vector< vector<real32> > &image, 
				  vector< vector<real32> > &exposureImage,
				  //vector< vector<int16> > &qualityImage, vector<string> &fileList);
				  vector< vector<int16> > &qualityImage, vector<string> &fileList,
				  double deadFracMax, double frameTimeMax);


    virtual void convertToLowResImage(vector< vector<float> > &image, 
                vector< vector<float> > &exposureImage,
		vector< vector<int16> > &qualityImage);






    virtual void correctReferencePixel(double dx, double dy);
    virtual void correctReferencePixel();

    virtual void updateReferencePixelCorrection(OMWCS &c, double dx, double dy);
 
    virtual double distance(double x1, double y1, OMWCS &c, double &x2, double &y2);

    virtual double overlapFraction(OMWCS &c2);



 
 

    
}; 
/*************************************************
*
*
*************************************************/


class WcsData2 : public WcsData
{
public:
    WcsData2() : WcsData() {};

    ~WcsData2() { };
    WcsData2& operator=(const WcsData2& rhs);

 
    virtual bool load(string );
    virtual void convert();
    //   virtual void setImageOffsets( OMWCS &, OMWCS & );
    virtual void setImageParameters();
    virtual void getCornerCoords();
    inline virtual void convertX1Y1ToX2Y2(double xIn, double yIn, OMWCS &c1, double &xOut, double &yOut);



    virtual bool withinImage(double x, double y);

    virtual void createOutputFile(string fileName, vector< vector<real32> > &image, 
				  vector< vector<real32> > &exposureImage,
				  //vector< vector<int16> > &qualityImage, vector<string> &fileList);
				  vector< vector<int16> > &qualityImage, vector<string> &fileList,
				  double  deadFracMax, double frameTimeMax);
    


    virtual void convertToLowResImage(vector< vector<float> > &image, 
                vector< vector<float> > &exposureImage,
		vector< vector<int16> > &qualityImage);







    virtual void correctReferencePixel(double dx, double dy);
    virtual void correctReferencePixel();

    virtual void updateReferencePixelCorrection(OMWCS &c, double dx, double dy);
 
    virtual double distance(double x1, double y1, OMWCS &c, double &x2, double &y2);

    virtual double overlapFraction(OMWCS &c2);

 
  
   
}; 


/***********************************************************
*
* The base class
*
***********************************************************/
class OMWCS
{


public:

    int imageType;  
    

    WcsData *data;

    OMWCS(int k);  
    OMWCS();

    OMWCS( const OMWCS &);
    OMWCS& operator=(const OMWCS &);

    ~OMWCS();


    /**************************************************************************
    *
    *
    **************************************************************************/

    bool  operator<(const OMWCS &a) const 
    {
     
     
      return data->_mjdMid < a.data->_mjdMid;
      

    };

    void setImageType(int type)
    {
        if(data) delete data;

        if(type != imageType)
        {
            imageType = type;
        }

        if (type == 0)
        {
            data = new WcsData1;
        }
        else
        {
            data = new WcsData2;
        }
    }  


};

/*******************************************************************
*
*
*
*******************************************************************/
void GetWindowParameters(DataSet *dataSetH, int &binAxis, bool &binBpe, 
			 double &windowX0, double &windowY0, int &nx, int &ny);    






void convertToCelestial(OMWCS &c, double x, double y, double &ra, double &dec);
inline void convertToCelestial1(OMWCS &c, double x, double y, double &ra, double &dec);
inline void convertToCelestial(WcsData1 &c, double x, double y, double &ra, double &dec);
inline void convertFromCelestial(double ra, double dec, OMWCS& c, double &x, double &y);

inline void removeDistotion(OMWCS &wcs, double x1, double y1, double x2, double y2);
inline void addDistotion(OMWCS &wcs, double x1, double y1, double x2, double y2);


/****************************************************************
*
* Convert x1, y1 to full frame coords
*
****************************************************************/
inline void imageToFullFrameCoords(double x, double y, double crpix1, double crpix2, 
			    int binAxis, double &x1, double &y2);



/**************************************************************
* Convert x, y from full frame coords
*
****************************************************************/
inline void fullFrameToImageCoords(double x, double y, double crpix1, double crpix2,
			    int binAxis, double &x1, double &y2);



inline void tangentPlaneToCelestial(WcsData1 &wcs, double xi, double eta, double &ra, double &dec);

inline void celestialToTangentPlane(WcsData1 & wcs, double ra, double dec, double &x, double &y);

#endif
