// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef UTILITY_H
#define UTILITY_H

#include "Dal.h"
#include <cmath>
#include<vector>
#include<iostream>
#include <cstdlib>
#include <cerrno>
#include<numeric>







double ModifiedJulianDate(const string dateString);





const double TWO_PI = 2.0 * M_PI;
const double radiansToDegrees = (double)180.0 / M_PI;
const double radiansToArcSecs = (double)(180 * 3600) / M_PI;



const double degreesToRadians = M_PI / (double)180.0;





double distance(double r1, double d1, double r2, double d2);

double Error(double mag1, double mag2);



vector<vector<int8> >   loadInt8Image(const DataSet *set, const char *name, int &nx, int &ny);
vector<vector<int16> >  loadInt16Image(const DataSet *set, const char *name, int &nx, int &ny);

vector<vector<float> >  loadFloatImage(const DataSet *set, const char *name, int &nx, int &ny);




vector<vector<float> >   loadFloatImage(const string set, const char *name, int &nx, int &ny);
vector<vector<int16> >   loadInt16Image(const string set, const char *name, int &nx, int &ny);







bool modifyInt16Image(DataSet *set, const char * name, const vector<vector<int16> > &image);





template<class kind>
void setBit(kind & value, int no)
{

    int l = sizeof(kind) * 8;
  
 
    value |= (1 << (no % l));


}
template<class kind>
void unsetBit(kind &value, int no)
{
    int l = sizeof(kind) * 8;
 
   
    value &= ~(1 << (no % l));


}
template<class kind>
bool bitSet(kind value, int no)
{

   int l = sizeof(value) * 8;
 

  return value& (1 << (no % l));

}

template<class T>
vector<T> difference(vector<T> &x, vector<T> &y, T constant=1);


template<class T>
vector<T> difference(vector<T> &x, vector<T> &y, T constant)
{
    vector<T> diff(x.size() );
    for(unsigned i = 0; i < x.size(); i++)
      diff[i] = (x[i] - y[i]) * constant;

    return diff;


}




/*************************************************************
*
* For determining if a value is NULL
*
****************************************************************/
bool isNull(float value);
bool isNull( double value);
bool isNull( int16 value);













/************************************************
*
*
************************************************/
template<class kind>
void smoothImage(vector<vector<kind> > &image)
{

    int nx = image.size() - 1;
    int ny = image[0].size() - 1;
    vector<vector<kind> > tempImage = image;

  
      

    double weight0 = 0.25;
    double weight1 = 1.0 / 8.0;
    double weight2 = 1.0 / 16.0;
  
    double sum = 0.0, sumWeight = 0.0;
    
    for(int i = 0; i <= nx; i++)
    {
        for(int j = 0; j <= ny; j++)
	{ 
	    
            if( !isNull( image[i][j] ) )
	    {
	        sum = image[i][j] * weight0;
                sumWeight = weight0;
	    }
            else
	    {
	        sum = 0.0;
                sumWeight = 0.0;
            }
              
            if (i > 0) 
	    { 
               
                if( !isNull( image[i - 1][j] ) ) 
		{
		    sum += double(image[i - 1][ j]) * weight1; // Adjacent left pixel
                    sumWeight += weight1;
                }
                if (j > 0) 
		{
                   
                    if( !isNull( image[i - 1][ j - 1] ) ) 
		    {
       
		        sum += double(image[i - 1][ j - 1]) * weight2; //  Adjacent bottom-left pixel
                        sumWeight += weight2;
                    }
                }
        
                if (j < ny) 
		{
                   
                    if( !isNull( image[i - 1][ j + 1] ) ) 
		    {
		        sum += double(image[i - 1][ j + 1]) * weight2; // Adjacent top-left pixel
                        sumWeight += weight2;
                    }
                }        

            }

            if (j > 0) 
	    {
      
                if( !isNull( image[i][ j - 1] ) ) 
		{
       
		    sum += double(image[i][ j - 1]) * weight1; // Adjacent bottom pixel
                    sumWeight += weight1;
                }
            }        
            if (j < ny) 
	    {
              
                if( !isNull( image[i][j + 1] ) ) 
		{
       
		    sum += double(image[i][ j + 1]) * weight1; // Adjacent top pixel
                    sumWeight += weight1;
                }
            }        

            if (i < nx) 
	    { 
                if( !isNull( image[i + 1][ j] ) ) 
		{
                 
		    sum += double(image[i + 1][ j]) * weight1; // Adjacent right pixel
                    sumWeight += weight1;
                }
                if (j > 0) 
		{
                    if( !isNull( image[i + 1][j - 1] ) ) 
		    {
       
		        sum += double( image[i + 1][ j - 1]) * weight2; // Adjacent bottom-right pixel
                        sumWeight += weight2;
                    }
                }        
                if (j < ny) 
		{
                  
                    if( !isNull( image[i + 1][ j + 1] ) ) 
		    {  
       
		        sum += double( image[i + 1][ j + 1]) * weight2;  // Adjacent top-right pixel
                        sumWeight += weight2;
                    }
                }        

            }
            if(sumWeight > 0.0) 
	    {
	        tempImage[i][j] = sum / sumWeight;
            }
	}
    }
    
    image = tempImage;
  
 
}


void addBlock(DataSet *set, string name, vector< vector<float> > &image, int position = 0);
void addBlock(DataSet *set, string name, vector< vector<double> > &image, int position = 0);
void addBlock(DataSet *set, string name, vector< vector<int16> > &image, int position = 0);


void normalizeImage(vector<vector<float > >& image1, vector<vector<float> > &image2);


#endif
